@@ -5,26 +5,44 @@ import json
 import logging
 from datetime import datetime, timedelta
 
+import asyncio
 import aiohttp
+import urllib.parse as parse
 
 MANUFACTURER = "Reolink"
 DEFAULT_STREAM = "main"
 DEFAULT_PROTOCOL = "rtmp"
 DEFAULT_CHANNEL = 0
+DEFAULT_TIMEOUT = 30
 
 _LOGGER = logging.getLogger(__name__)
 
 
-class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public-methods
+class Api:  # pylint: disable=too-many-instance-attributes disable=too-many-public-methods
     """Reolink API class."""
 
-    def __init__(self, host, port, username, password):
+    def __init__(
+        self,
+        host,
+        port,
+        username,
+        password,
+        channel=DEFAULT_CHANNEL,
+        protocol=DEFAULT_PROTOCOL,
+        stream=DEFAULT_STREAM,
+        timeout=DEFAULT_TIMEOUT,
+    ):
         """Initialize the API class."""
         self._url = f"http://{host}:{port}/cgi-bin/api.cgi"
         self._host = host
         self._port = port
         self._username = username
         self._password = password
+        self._channel = channel
+        self._stream = stream
+        self._protocol = protocol
+        self._timeout = aiohttp.ClientTimeout(total=timeout)
+
         self._token = None
         self._lease_time = None
         self._motion_state = False
@@ -39,8 +57,14 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
         self._rtsp_port = None
         self._rtmp_port = None
         self._onvifport = None
+        self._mac_address = None
+        self._serial = None
+        self._name = None
+        self._sw_version = None
+        self._model = None
+        self._channels = None
         self._ptz_presets = dict()
-        self._sensititivy_presets = dict()
+        self._sensitivity_presets = dict()
         self._motion_detection_state = None
 
         self._isp_settings = None
@@ -56,9 +80,6 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
 
         self._users = None
         self._local_link = None
-        self._stream = DEFAULT_STREAM
-        self._protocol = DEFAULT_PROTOCOL
-        self._channel = DEFAULT_CHANNEL
         self._ptz_support = False
 
     @property
@@ -79,33 +100,38 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
     @property
     def mac_address(self):
         """Return the mac address."""
-        return self._local_link["value"]["LocalLink"]["mac"]
+        return self._mac_address
 
     @property
     def serial(self):
         """Return the serial."""
-        return self._device_info["value"]["DevInfo"]["serial"]
+        return self._serial
 
     @property
     def name(self):
         """Return the camera name."""
-        return self._device_info["value"]["DevInfo"]["name"]
+        return self._name
 
     @property
     def sw_version(self):
         """Return the software version."""
-        return self._device_info["value"]["DevInfo"]["firmVer"]
+        return self._sw_version
 
     @property
     def model(self):
         """Return the model."""
-        return self._device_info["value"]["DevInfo"]["model"]
+        return self._model
 
     @property
     def manufacturer(self):
         """Return the manufacturer name (Reolink)."""
         return MANUFACTURER
 
+    @property
+    def channels(self):
+        """Return the number of channels."""
+        return self._channels
+
     @property
     def motion_state(self):
         """Return the motion state (polling)."""
@@ -157,9 +183,9 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
         return self._ptz_presets
 
     @property
-    def sensititivy_presets(self):
+    def sensitivity_presets(self):
         """Return the sensitivity presets."""
-        return self._sensititivy_presets
+        return self._sensitivity_presets
 
     @property
     def device_info(self):
@@ -242,8 +268,8 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
         if len(self._ptz_presets) != 0:
             capabilities.append("ptzPresets")
 
-        if len(self._sensititivy_presets) != 0:
-            capabilities.append("sensititivyPresets")
+        if len(self._sensitivity_presets) != 0:
+            capabilities.append("sensitivityPresets")
 
         return capabilities
 
@@ -257,6 +283,7 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
             {"cmd": "GetIrLights", "action": 1, "param": {"channel": self._channel}},
             {"cmd": "GetRec", "action": 1, "param": {"channel": self._channel}},
             {"cmd": "GetPtzPreset", "action": 1, "param": {"channel": self._channel}},
+            {"cmd": "GetHddInfo", "action": 1, "param": {}},
             {
                 "cmd": "GetAlarm",
                 "action": 1,
@@ -270,12 +297,17 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
                     body.pop(x)
 
         response = await self.send(body)
+        if response is None:
+            return False
+
         try:
             json_data = json.loads(response)
             await self.map_json_response(json_data)
             return True
         except (TypeError, json.JSONDecodeError):
-            _LOGGER.error("Error translating Reolink state response")
+            _LOGGER.debug(
+                "Host: %s: Error translating Reolink state response", self._host
+            )
             await self.clear_token()
             return False
 
@@ -286,7 +318,7 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
             {"cmd": "GetLocalLink", "action": 1, "param": {"channel": self._channel}},
             {"cmd": "GetNetPort", "action": 1, "param": {"channel": self._channel}},
             {"cmd": "GetUser", "action": 1, "param": {"channel": self._channel}},
-            {"cmd": "GetHddInfo", "action": 1, "param": {}},
+            {"cmd": "GetOsd", "action": 1, "param": {"channel": self._channel}},
             {
                 "cmd": "GetAbility",
                 "action": 1,
@@ -295,13 +327,17 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
         ]
 
         response = await self.send(body)
+        if response is None:
+            return False
 
         try:
             json_data = json.loads(response)
             await self.map_json_response(json_data)
             return True
         except (TypeError, json.JSONDecodeError):
-            _LOGGER.error("Error translating Reolink settings response")
+            _LOGGER.debug(
+                "Host %s: Error translating Reolink settings response", self._host
+            )
             await self.clear_token()
             return False
 
@@ -310,14 +346,15 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
         body = [{"cmd": "GetMdState", "action": 0, "param": {"channel": self._channel}}]
 
         response = await self.send(body)
+        if response is None:
+            return False
 
         try:
             json_data = json.loads(response)
 
             if json_data is None:
                 _LOGGER.error(
-                    "Unable to get Motion detection state at IP %s",
-                    self._host
+                    "Unable to get Motion detection state at IP %s", self._host
                 )
                 self._motion_state = False
                 return self._motion_state
@@ -332,10 +369,11 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
     async def get_still_image(self):
         """Get the still image."""
         param = {"cmd": "Snap", "channel": self._channel}
-        response = await self.send(None, param)
 
-        if response is None or response == b'':
+        response = await self.send(None, param)
+        if response is None or response == b"":
             return
+
         return response
 
     async def get_snapshot(self):
@@ -350,18 +388,13 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
         if self.protocol == DEFAULT_PROTOCOL:
             stream_source = f"rtmp://{self._host}:{self._rtmp_port}/bcs/channel{self._channel}_{self._stream}.bcs?channel={self._channel}&stream=0&token={self._token}"
         else:
-            channel = "{:02d}".format(self._channel+1)
-            stream_source = f"rtsp://{self._host}:{self._rtsp_port}/h264Preview_{channel}_{self._stream}&token={self._token}"
+            password = parse.quote(self._password)
+            channel = "{:02d}".format(self._channel + 1)
+            stream_source = f"rtsp://{self._username}:{password}@{self._host}:{self._rtsp_port}/h264Preview_{channel}_{self._stream}"
 
         return stream_source
 
-    async def update_streaming_options(self, stream, protocol, channel):
-        """Update the streaming options."""
-        self._stream = stream
-        self._protocol = protocol
-        self._channel = channel
-
-    async def map_json_response(self, json_data): #pylint: disable=too-many-branches
+    async def map_json_response(self, json_data):  # pylint: disable=too-many-branches
         """Map the JSON objects to internal objects and store for later use."""
         for data in json_data:
             try:
@@ -370,12 +403,18 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
 
                 if data["cmd"] == "GetDevInfo":
                     self._device_info = data
+                    self._serial = data["value"]["DevInfo"]["serial"]
+                    self._device_name = data["value"]["DevInfo"]["name"]
+                    self._sw_version = data["value"]["DevInfo"]["firmVer"]
+                    self._model = data["value"]["DevInfo"]["model"]
+                    self._channels = data["value"]["DevInfo"]["channelNum"]
 
-                if data["cmd"] == "GetHddInfo":
+                elif data["cmd"] == "GetHddInfo":
                     self._hdd_info = data
 
-                if data["cmd"] == "GetLocalLink":
+                elif data["cmd"] == "GetLocalLink":
                     self._local_link = data
+                    self._mac_address = data["value"]["LocalLink"]["mac"]
 
                 elif data["cmd"] == "GetNetPort":
                     self._netport_settings = data
@@ -383,7 +422,11 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
                     self._rtmp_port = data["value"]["NetPort"]["rtmpPort"]
                     self._onvifport = data["value"]["NetPort"]["onvifPort"]
 
-                if data["cmd"] == "GetUser":
+                elif data["cmd"] == "GetOsd":
+                    self._osd_settings = data
+                    self._name = data["value"]["Osd"]["osdChannel"]["name"]
+
+                elif data["cmd"] == "GetUser":
                     self._users = data["value"]["User"]
 
                 elif data["cmd"] == "GetFtp":
@@ -402,7 +445,7 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
 
                 elif data["cmd"] == "GetIsp":
                     self._isp_settings = data
-                    self._daynight_state = data["value"]["Isp"]["dayNight"] == "Auto"
+                    self._daynight_state = data["value"]["Isp"]["dayNight"]
 
                 elif data["cmd"] == "GetIrLights":
                     self._ir_settings = data
@@ -425,7 +468,7 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
                 elif data["cmd"] == "GetAlarm":
                     self._alarm_settings = data
                     self._motion_detection_state = data["value"]["Alarm"]["enable"] == 1
-                    self._sensititivy_presets = data["value"]["Alarm"]["sens"]
+                    self._sensitivity_presets = data["value"]["Alarm"]["sens"]
 
                 elif data["cmd"] == "GetMdState":
                     self._motion_state = json_data[0]["value"]["state"] == 1
@@ -433,7 +476,7 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
                 elif data["cmd"] == "GetAbility":
                     for ability in data["value"]["Ability"]["abilityChn"]:
                         self._ptz_support = ability["ptzCtrl"]["permit"] != 0
-            except: #pylint: disable=bare-except
+            except:  # pylint: disable=bare-except
                 continue
 
     async def login(self):
@@ -443,7 +486,9 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
 
         _LOGGER.debug(
             "Reolink camera with host %s:%s trying to login with user %s",
-            self._host, self._port, self._username
+            self._host,
+            self._port,
+            self._username,
         )
 
         body = [
@@ -451,19 +496,26 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
                 "cmd": "Login",
                 "action": 0,
                 "param": {
-                    "User": {"userName": self._username, "password": self._password}
+                    "User": {
+                        "userName": self._username,
+                        "password": self._password[:31],
+                    }
                 },
             }
         ]
         param = {"cmd": "Login", "token": "null"}
 
         response = await self.send(body, param)
+        if response is None:
+            return False
 
         try:
             json_data = json.loads(response)
             _LOGGER.debug("Get response from %s: %s", self._host, json_data)
         except (TypeError, json.JSONDecodeError):
-            _LOGGER.error("Error translating login response to json")
+            _LOGGER.debug(
+                "Host %s: Error translating login response to json", self._host
+            )
             return False
 
         if json_data is not None:
@@ -474,11 +526,13 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
 
                 _LOGGER.debug(
                     "Reolink camera logged in at IP %s. Leasetime %s, token %s",
-                    self._host, self._lease_time.strftime('%d-%m-%Y %H:%M'), self._token
+                    self._host,
+                    self._lease_time.strftime("%d-%m-%Y %H:%M"),
+                    self._token,
                 )
                 return True
 
-        _LOGGER.error("Failed to login at IP %s. Connection error.", self._host)
+        _LOGGER.debug("Failed to login at IP %s.", self._host)
         return False
 
     async def is_admin(self):
@@ -488,13 +542,15 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
                 if user["level"] == "admin":
                     _LOGGER.debug(
                         "User %s has authorisation level %s",
-                        self._username, user['level']
+                        self._username,
+                        user["level"],
                     )
                 else:
                     _LOGGER.warning(
                         """User %s has authorisation level %s. Only admin users can change
                         camera settings! Switches will not work.""",
-                        self._username, user['level']
+                        self._username,
+                        user["level"],
                     )
 
     async def logout(self):
@@ -505,6 +561,22 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
         await self.send(body, param)
         await self.clear_token()
 
+    async def set_channel(self, channel):
+        """Update the channel property."""
+        self._channel = channel
+
+    async def set_stream(self, stream):
+        """Update the stream property."""
+        self._stream = stream
+
+    async def set_protocol(self, protocol):
+        """Update the protocol property."""
+        self._protocol = protocol
+
+    async def set_timeout(self, timeout):
+        """Update the timeout property."""
+        self._timeout = aiohttp.ClientTimeout(total=timeout)
+
     async def set_ftp(self, enable):
         """Set the FTP parameter."""
         if not self._ftp_settings:
@@ -630,7 +702,7 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
         return await self.send_setting(body)
 
     async def set_sensitivity(self, value: int, preset=None):
-        """Set motion detection sensititivy.
+        """Set motion detection sensitivity.
         Here the camera web and windows application
         show a completely different value than set.
         So the calculation 51-value makes the "real"
@@ -644,15 +716,13 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
             {
                 "cmd": "SetAlarm",
                 "action": 1,
-                "param": {	
+                "param": {
                     "Alarm": {
                         "channel": 0,
                         "type": "md",
-                        "sens": 
-                            self._alarm_settings["value"]["Alarm"]["sens"],
-                        
+                        "sens": self._alarm_settings["value"]["Alarm"]["sens"],
                     }
-                }
+                },
             }
         ]
         for setting in body[0]["param"]["Alarm"]["sens"]:
@@ -662,7 +732,7 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
         return await self.send_setting(body)
 
     async def set_ptz_command(self, command, preset=None, speed=None):
-        '''Send PTZ command to the camera.
+        """Send PTZ command to the camera.
 
         List of possible commands
         --------------------------
@@ -683,7 +753,7 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
         ToPos       X       X
         Auto
         Stop
-        '''
+        """
 
         body = [
             {
@@ -703,10 +773,12 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
         """Send a setting."""
         command = body[0]["cmd"]
         _LOGGER.debug(
-            "Sending command: %s to: %s with body: %s",
-            command, self._host, body
+            "Sending command: %s to: %s with body: %s", command, self._host, body
         )
         response = await self.send(body, {"cmd": command})
+        if response is None:
+            return False
+
         try:
             json_data = json.loads(response)
             _LOGGER.debug("Response from %s: %s", self._host, json_data)
@@ -718,10 +790,16 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
 
             return False
         except (TypeError, json.JSONDecodeError):
-            _LOGGER.error("Error translating %s response to json", command)
+            _LOGGER.debug(
+                "Host %s: Error translating %s response to json", self._host, command
+            )
             return False
         except KeyError:
-            _LOGGER.error("Received an unexpected response while sending command: %s", command)
+            _LOGGER.debug(
+                "Host %s: Received an unexpected response while sending command: %s",
+                self._host,
+                command,
+            )
             return False
 
     async def send(self, body, param=None):
@@ -731,23 +809,30 @@ class Api: #pylint: disable=too-many-instance-attributes disable=too-many-public
                 return False
 
         if not param:
-            param={}
+            param = {}
         if self._token is not None:
             param["token"] = self._token
 
-        timeout = aiohttp.ClientTimeout(total=10)
-
         try:
             if body is None:
-                async with aiohttp.ClientSession(timeout=timeout) as session:
+                async with aiohttp.ClientSession(timeout=self._timeout) as session:
                     async with session.get(url=self._url, params=param) as response:
                         return await response.read()
             else:
-                async with aiohttp.ClientSession(timeout=timeout) as session:
+                async with aiohttp.ClientSession(timeout=self._timeout) as session:
                     async with session.post(
                         url=self._url, json=body, params=param
                     ) as response:
                         json_data = await response.text()
                         return json_data
-        except: #pylint: disable=bare-except
-            return ""
+
+        except aiohttp.ClientConnectorError as conn_err:
+            _LOGGER.debug("Host %s: Connection error %s", self._host, str(conn_err))
+        except asyncio.TimeoutError:
+            _LOGGER.debug(
+                "Host %s: connection timeout exception. Please check the connection to this camera.",
+                self._host,
+            )
+        except:  # pylint: disable=bare-except
+            _LOGGER.debug("Host %s: Unknown exception occurred.", self._host)
+        return
